// gcc -o test test.c -lm
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define IO_TIME 30

typedef struct job {
	int id;            
	int arrival;       
	int unblock; 
	int turnaround;
	int reaction;
	int exectime;      
	float ioratio;     
	struct job *next;
} job;

typedef struct spec {
	int arrival;
	int exectime;
	float ioratio;
} spec;

spec specs[] = {
	{0, 10, 0.0},
	{0, 30, 0.7},
	{0, 20, 0.0},
	{40, 80, 0.4},
	{60, 30, 0.3},  
	{120, 90, 0.3},  
	{120, 40, 0.5},
	{140, 20, 0.2},
	{160, 10, 0.3},
	{180, 20, 0.3},
	{0, 0, 0}
};

job *readyq = NULL;
job *blockedq = NULL;
job *doneq = NULL;

void ready(job *this) {
	job *nxt = readyq;
	job *prev = NULL;
	while (nxt != NULL ) {
		prev = nxt;
		nxt = nxt->next;
	}

	this->next = nxt;
	if (prev == NULL) {
		readyq = this;
	} else {
		prev->next = this;
	}

	return;
}

void block(job *this) {
	job *nxt = blockedq;
	job *prev = NULL;

	while (nxt != NULL) {
		if (this->unblock < nxt->unblock) {
			break;
		} else {
			prev = nxt;
			nxt = nxt->next;
		}
	}

	this->next = nxt;
	if (prev != NULL) {
		prev->next = this;
	} else {
		blockedq = this;
	}

	return;
}

void unblock(int time) {
	while (blockedq != NULL && blockedq->unblock <= time) {
		job *nxt = blockedq;
		blockedq = nxt->next; 
		ready(nxt);
	}
}

void done(job *this) {
	this->next = doneq;
	doneq = this;

	return;
}

void init() {
	int i = 0;
	while (specs[i].exectime != 0) {
		job *new = (job *)malloc(sizeof(job));
		new->id = i+1;
		new->arrival = specs[i].arrival;
		new->unblock = specs[i].arrival;
		new->exectime = specs[i].exectime;
		new->ioratio = specs[i].ioratio;
		new->turnaround = 0;
		new->reaction = 0;
		block(new);
		i++;
	}
}

int io_op(float ratio) {
	return ((float)rand()) / RAND_MAX < ratio;
}


int schedule(int time, int slot) {
	if (readyq != NULL) {
		job *nxt = readyq;
    	readyq = readyq->next;

		if (nxt->reaction == 0) 
			nxt->reaction = time - nxt->arrival;

		int left = nxt->exectime;
		int io = 0;
		int exect = (left < slot) ? left : slot;

		if (exect > 1) {
			io = io_op(nxt->ioratio);
			io = 0;
			if (io) {
				exect = (int)trunc(((float)rand())/RAND_MAX * (exect - 1)) + 1;
			}
		}

		nxt->exectime -= exect;
		printf("(%4d) scheduling job %2d for %3d ms ", time, nxt->id, exect);

		if (nxt->exectime == 0) {
			nxt->turnaround = time + exect - nxt->arrival;
			printf("- done\n");
			done(nxt);
		} else {
			if (io) {
				nxt->unblock = time + exect + IO_TIME;
				block(nxt);
				printf("- %3d left - blocked\n", nxt->exectime);
			} else {
				ready(nxt);
				printf("- %3d left\n", nxt->exectime);	
			}
		}

		return exect;
	} else {
		return 1;
	}
}

int main(int argc, char *argv[]) {
	int slot = 10;
  
	if (argc == 2) {
		slot = atoi(argv[1]);
	}

	printf("slot = %d\n", slot);
	init();

	int time = 0;
	while (blockedq != NULL || readyq != NULL) {
		unblock(time);
		int tick = schedule(time, slot);
		time += tick;
	}

	int turnaround = 0;
	int reaction = 0;
	int jobs = 0;  
  
	for (job *nxt = doneq; nxt != NULL; nxt = nxt->next) {
		jobs += 1;
		turnaround += nxt->turnaround;
		reaction += nxt->reaction;
	}


	printf("\naverage turnaround: %d \n", (int)trunc((float)turnaround/jobs));
	printf("\naverage reaction: %d \n", (int)trunc((float)reaction/jobs));    
	printf("\ntotal execution time is %d \n", time);
  
	return 0;
}

